/*   BEM = Batch Equilibrium Model (calculates chemical equilibira)
     Copyright (C) 1996  J. Prenzel (jprenze@gwdg.de)
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.         */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <time.h>
#include <sys/stat.h>

#include "bem.h"


/*--------------------------------------- bemvrs -------------------------*/
char *bemvrs(){
/*  return "bem-6.00";  bem in C, code completely new, started Oct 1995 */
/*  return "bem-6.01t";    started 14.12.95 */
  return "bem-6.02";   /* started 25.01.95 */
}

/*--------------------------------------- hdgres -------------------------*/
int hdgres(){
  time_t zeit;
  time(&zeit);
  fprintf(Res.fil,"# Results from %s. %s# Files used "
          "(with time of last modification):\n",bemvrs(),ctime(&zeit));
  if(Op.s)fprintf(Sfil.fil,"Results from %s. %sFiles used "
          "(with time of last modification):\n",bemvrs(),ctime(&zeit));
  return 0;
}

/*--------------------------------------- recfil -------------------------*/
/* identify an input file with modification date to  output files */
int recfil(FIL fil){
  struct stat s;
  int n;
  time_t t;
  n=fileno(fil.fil);
  fstat(n,&s);
  t=s.st_mtime;
  fprintf(Res.fil,"# %s  %s",fil.nam,ctime(&t));
  if(Op.s)fprintf(Sfil.fil,"%s  %s",fil.nam,ctime(&t));
  return 0;
}

/*....................................... openfiles ..............*/
int openfiles(char typ){

  switch(typ){
  case 'f':
    fclose(Flowfil);
    Flowfil=fopen("bem01.tmp","w+");
    if(Flowfil == NULL){
      bemerr(1,"bem01.tmp","","",errno);
      return -32;
    }
    break;
  case 'g':
    Putsfil=fopen("bem02.tmp","w+");
    if(Putsfil == NULL){
      bemerr(1,"bem02.tmp","","",errno);
      return -32;
    }
    break;
 case 'm':
    Mod.fil=fopen(Mod.nam,"r");
    if(Mod.fil == NULL){
      bemerr(1,Mod.nam,"","",errno);
      return -32;
    }
    break;
  case 'i':
    Ifil.fil=fopen(Ifil.nam,"wb");
    if(Ifil.fil == NULL){
      bemerr(1,Ifil.nam,"","",0);
      return -32;
    }
    break;
  case 'q':
    Qfil.fil=fopen(Qfil.nam,"rb");
    if(Qfil.fil == NULL){
      bemerr(1,Qfil.nam,"","",0);
      return -32;
    }
    break;
  case 'r':
    Res.fil=fopen(Res.nam,"w");
    if(Res.fil == NULL){
      bemerr(1,Res.nam,"","",errno);
      return -32;
    }
    break;
  case 's':
    if(strcmp(Sfil.nam,"-") == 0){
      Sfil.fil = stdout;
    }
    else{
      Sfil.fil=fopen(Sfil.nam,"w");
    }
    if(Sfil.fil == NULL){
      bemerr(1,Sfil.nam,"","",errno);
      return -32;
    }
    break;
  case 't':
    Tat.fil=fopen(Tat.nam,"r");
    if(Tat.fil == NULL){
      bemerr(1,Tat.nam,"","",errno);
      return -32;
    }
    break;
  case 'l':
    Lgf.nam=getenv("BEMLOG");
    if(Lgf.nam == NULL || Lgf.nam[0] == '-'){
      Lgf.fil=stdout;
    }
    else{
      Lgf.fil=fopen(Lgf.nam,"w");
      if(Lgf.fil == NULL){
	bemerr(1,Lgf.nam,"($Lgf)","",errno);
	return -32;
      }
    }
  }

  return 0;
}

/*............................................... bemerr ...........*/
void bemerr( int ierr, char *obj1, char *obj2, char *obj3, int ier){
  FILE *errout;
  errout=stderr;

  fprintf(errout,"(%s:) ",Actfil);
  switch (ierr){
  case 1:
    fprintf(errout,"BEM error: Open error for file \"%s\" %s.\n",obj1,obj2);
    break;
  case 2:
    fprintf(errout,"BEM error: Too many words in record # %s, file %s\n",obj1,obj2);
    break;
  case 3:
    fprintf(errout,"BEM error: Not enough memory for element %s\n",obj1);
    break;
  case 4:
    fprintf(errout,"BEM error: too many BEM - elements.\n");
    break;
  case 5:
    fprintf(errout,"BEM error: too many BEM - complexes\n");
    break;
  case 6:
    fprintf(errout,"BEM error: Not enough memory for more complexes .\n");
    break;
  case 7:
    fprintf(errout,"BEM error: Invalid definition: %s\n",obj1);
    break;
  case 8:
    fprintf(errout,"BEM error: Invalid number \"%s\" for %s (errno=%d)\n",obj1,obj2,ier);
    break;
  case 9:
    fprintf(errout,"BEM error: Invalid number \"%s\" in %s (errno=%d)\n",obj1,obj2,ier);
    break;
  case 10:
    fprintf(errout,"BEM error: Unknown Element \"%s\" in definition %s\n",obj1,obj2);
    break;
  case 11:
    fprintf(errout,"BEM error: Charge \"%s\" for %s too large \n",obj1,obj2);
    break;
  case 12:
    fprintf(errout,"BEM error: Not enough memory for reaction cell %s\n",obj1);
    break;
  case 13:
    fprintf(errout,"BEM error: unknown statement type: %s %s %s ...\n",
           obj1,obj2,obj3);
    break;
  case 14:
    fprintf(errout,"BEM error: end of statement - no value for %s ?\n",obj1);
    break;
  case 15:
    fprintf(errout,"BEM error: unknown keyword \"%s\" in %s statement.\n",obj1,obj2);
    break;
  case 16:
    fprintf(errout,"BEM error: you must supply a value for %s.\n",obj1);
    break;
  case 17:
    fprintf(errout,"BEM warning: no value for %s.  Assumed: %s\n",obj1,obj2);
    break;
  case 18:
    fprintf(errout,"BEM error: total concentration is zero for %s.\n",obj1);
    break;
  case 19:
    fprintf(errout,"BEM warning: no value for %s of %s. Assumed: %s\n"
            ,obj1,obj2,obj3);
    break;
  case 20:
    fprintf(errout,"BEM error: too many BEM - salts\n");
    break;
  case 21:
    fprintf(errout,"BEM error: Not enough memory for more salts .\n");
    break;
  case 22:
    fprintf(errout,"BEM error: Ionic charge %s for salt %s is impossible. \n"
            ,obj1,obj2);
    break;
  case 23:
    fprintf(errout,"BEM error: Unknown command line argument \"%s\". \n",obj1);
    break;
  case 24:
    fprintf(errout,"BEM error: Definition \"%s\" makes no sense because %s is not defined.\n",obj1,obj2);
    break;
  case 25:
    fprintf(errout,"BEM error: Not enough memory for more %s .\n",obj1);
    break;
  case 26:
    fprintf(errout,"BEM error: too many BEM - cation exchangers.\n");
    break;
  case 27:
    fprintf(errout,"BEM warning: Insufficient cation exchange selectivity information for %s. Assumed pKsel=0.\n",obj1);
    break;
  case 28:
    fprintf(errout,"BEM error: incompatible specifications for cation exchanger.\nUse either direct (xs0 / xpksel) or indirect (xeql / xeqs) method.\n");
    break;
  case 29:
    fprintf(errout,"BEM error: Too many words in record # %s\n",obj1);
    break;
  case 30:
    fprintf(errout,"BEM error: Errors in first trial to calculate system. ier =%d",ier);
    break;
  case 31:
    fprintf(errout,"BEM error: Too few words in %s - line.\n",obj1 );
    break;
  case 32:
    fprintf(errout,"BEM error: Statement out of order: %s %s %s ...\n",obj1
            ,obj2,obj3);
    break;
  case 33:
    fprintf(errout,"BEM error: Item too long: %s.\n",obj1);
    break;
  case 34:
    fprintf(errout,"BEM error: Cell \"%s\"not found.\n",obj1);
    break;
  case 35:
    fprintf(errout,"BEM error: Internal file Flowfil corrupted.\n");
    break;
  case 36:
    fprintf(errout,"BEM error: Reaction cell %s can not deal with "
            "input of %s.\n",obj1,obj2);
    break;
  case 37:
    fprintf(errout,"BEM error: Unknown command line option \"%s\". \n",obj1);
    break;
  case 38:
    fprintf(errout,"BEM error: Command line option \"%s\" requires an argument. \n",obj1);
    break;
  case 39:
    fprintf(errout,"BEM failure: No convergence after %d function calls. \n",ier);
    break;
  case 40:
    fprintf(errout,"BEM failure: No starting equilibrium found. \n");
    break;
  case 41:
    fprintf(errout,"BEM error: File %s corrupt. \n",obj1);
    break;
  case 42:
    fprintf(errout,"BEM error: %s becomes negative at cell %s step %d \n",obj1,obj2,ier);
    break;
  case 43:
    fprintf(errout,"BEM error: Unknown type of solubility product function \"%s\"\n",obj1);
    break;
  case 44:
    fprintf(errout,"BEM error: Too many parameters for solubility product function for %s\n",obj1);
    break;
  default:
    fprintf(errout,"%d is an unknown error number!\n",ierr);
  }
  return;
}



/*............................................... copnb .............*/

/* copies n bytes from string a to string b and closes string b with a
 terminal null. If n > max-1, "?" is copied to b*/

void copnb(char *a, char *b, int n, int max){
  if(n<max){
    strncpy(b,a,n);
  }
  else{
    n=1;
    *b='?';
  }
  *(b+n)=0;
  return;
}


/*................................................ words .............*/
/* The function word takes the string rec and notes the beginning of
   space separated words in wrdpnt[]. The end of each  word is marked by
   a null character in rec itself! */

int words(char *rec, char *wrdpnt[]){
  int nwords,kword;
  size_t nchar;
  char *cpt;
  nwords=0;
  kword=0;
  nchar=strlen(rec);
  for(cpt=rec;cpt<rec+nchar;cpt++){
    if(kword==0){
      if(*cpt != ' ' && *cpt != '\t' && *cpt != '\n'){
	kword=1;
	nwords++;
	if(nwords>MAXWORDS){
	  return -12;
	}
	wrdpnt[nwords-1]=cpt;
      }
    }
    if(kword==1){
      if(*cpt == ' ' || *cpt == '\t' || *cpt == '\n'){
	kword=0;
	*cpt=0;
      }
    }
  }
  return nwords;
}

/*................................................. xlist1 ...............*/
int xlist1( double x[],int n){
  int i;
  for(i=0;i<n;i++){
    printf("%10g ",x[i]);
  }
  printf("\n");
  return 0;
}



/*------------------------------------------------- itrep ----------------*/
/* iteration report */
int itrep(int typ, int i1, int i2, int i3, double x1, double x2, double x3){
  int i;
  char form0[]="%s %3d %7d %7d %#9.4g %#9.4g "; 
  char form1[]=" ...  %#9.4g\n"; 
  char form2[]=" ==>  %#9.4g  %2d\n"; 
  char form3[]=" ==>  %#9.4g  %2d / %5d\n"; 

  fprintf(Lgf.fil,form0,Cl->id,Cl->step,Cl->it2cnt,Cl->it1cnt,Cl->alf
	  ,Cl->dalf);
  for(i=0;i<Cl->nslt;i++){
    fprintf(Lgf.fil,"%c",Cl->sl[i]->state);
  }

  switch (typ){
    
  case 1:
    fprintf(Lgf.fil,form1,x1);
    break;

  case 2:
    fprintf(Lgf.fil,form2,x1,i1);
    break;
    
  case 3:
    fprintf(Lgf.fil,form3,x1,i1,i2);
    break;
    
  }


  return 0;
}


/*------------------------------------------ steprdr ---------------------*/
/* This function reads from a textfile and repeats blocks that are
   started by "steps <nn>" where nn is the repeat count. Returns: 0 on
   normal read, 1 on end of block, 2 on end of file */

int steprdr(FILE *inp, char rec[]){
  char word[8],numb[8],*tail;
  int eof,eob;
  static int nnew=-999, nold=0;
  static fpos_t posnew,posold;

    fgets(rec,128,inp);
    if(feof(inp) == 0){
      eof=0;
      eob=0;
      sscanf(rec,"%8s%8s",word,numb);
      if(strcmp(word,"steps") == 0){
	if(nnew == -999){
	  errno=0;
	  nnew=strtod(numb,&tail);
	  if(errno != 0 || *tail != 0){
	    bemerr(8,numb,"steps","",errno);
	    return -32;
	  }
	  fgetpos(inp,&posnew);
	}
	eob=1;
      }
    }
    else{
      eob=1;
      eof=1;
    }
    if(eob > 0){
      if(nold <= 0){
	if(eof >0) return 2;
	nold=nnew;
	posold=posnew;
	nnew=-999;
      }
      if(nold  > 0){
	fsetpos(inp,&posold);
	nold--;
      }

    }
  return eob;
}

/*------------------------------------------------ dnum ----------------*/
int dnum(char *s, char *r, double *x){
  char *tail;
  errno=0;
  *x=strtod(s,&tail);
  if(strlen(tail)>0 || errno !=0) {
    bemerr(8,s,r,"",errno);
    return -32;
  }
  return 0;
}
/*------------------------------------------------- nxtkv --------------*/
/* get next key/value pair from file. Items space separated. Value = double.
   returns:
           0   success
           1   end of line
           2   end of file
          -1   key too long
          -2   error in number interpretation
*/
int nxtkv(FILE *inp, const int mkey, char *key, double *val ){
  int k=-999,lkey=-999,lnum=-999;
  char c,num[24],*tail;

  k=0;                          /* outside (key, value) */
  while(1){
    c=fgetc(inp);

    if(k==0 && c==EOF) return 2;
    if(k==0 && c=='\n') return 1;

    if(c != ' ' && c != '\t' && c != '\n' && c != EOF){
      if(k == 0){
	k=1;                    /* inside key */
	lkey=1;
	key[0]=c;
	continue;
      }
      if(k==1){
	lkey++;
	if(lkey>mkey) return -1;
	key[lkey-1]=c;
	continue;
      }
      if(k==2){
	k=3;                   /* inside value */
	lnum=1;
	num[0]=c;
	continue;
      }
      if(k==3){
	lnum++;
	if(lnum>24  ) return -2;
	num[lnum-1]=c;
	continue;
      }
     }
    else{
      if(k==0){
	continue;
      }
      if(k==1){
	k=2;                  /* between key and value */
	lkey++;
	if(lkey>mkey) return -1;
	key[lkey-1]=0;
	continue;
      }
      if(k==2){
	continue;
      }
      if(k==3){
	k=0;
	lnum++;
	if(lnum>24  ) return -2;
	num[lnum-1]=0;
	errno=0;
	*val=strtod(num,&tail);
	if(strlen(tail)>0 || errno != 0) return -2;
	ungetc(c,inp);
	return 0;
      }
    }
  }
  return 0;
}

/*-------------------------------------------- prpout -------------------*/
/* prepare data after equilibirum calculation: 
    1.) cttl   (total concentration in liquid phase, mol/L
    2.) csltt  (total concentration of salt: internal plus offset, mmol/kg)
    3.) pact   (-log10 of single ion activities in solution)        */
int prpout(CLL *cl0){
  int i,ii,iii,m;
  double a,c,d,s,p;
  double pgam[MAXCHRG+1];

  for(i=2;i<cl0->nelt;i++){
    cl0->el[i]->cttl=cl0->el[i]->csng;
    for(ii=0;ii<cl0->ncpl;ii++){
      for(iii=0;iii<cl0->cp[ii]->nelt;iii++){
	if(cl0->cp[ii]->el[iii]->num == i){
	cl0->el[i]->cttl += cl0->cp[ii]->ccpl * cl0->cp[ii]->stcf[iii];
	}
      }
    }
  }

  for(i=0;i<cl0->nslt;i++){
    cl0->sl[i]->csltt = cl0->sl[i]->cslt + cl0->sl[i]->coff;
  }

  d=cl0->ionstr;
  c=sqrt(d);
  a=0.5*(c/(1.0+c) - 0.3*d); /* Davies' equation used here */
  for(i=0;i<=MAXCHRG;i++){
    pgam[i]=a*i*i;
  }

  for(i=0;i<cl0->nelt;i++){
    s=cl0->el[i]->csng;
    m=abs(cl0->el[i]->chrg);
    p=-log(s)/M_LN10 + pgam[m];
    cl0->el[i]->pact=p;
  }


  return 0;
}
